"""Tiingo Helpers Module."""

from typing import Optional, Union


async def get_data(url: str) -> Union[dict, list]:
    """Get data from Tiingo endpoint and parse the JSON response."""
    # pylint: disable=import-outside-toplevel
    from json import JSONDecodeError  # noqa
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
    from openbb_core.provider.utils.helpers import amake_request

    response: Optional[Union[dict, list]] = None

    try:
        response = await amake_request(url)
        if (
            response
            and isinstance(response, dict)
            and len(list(response.keys())) == 1
            and response.get("detail")
        ):
            if (
                "token" in response["detail"]
                or "access" in response["detail"]
                or "permission" in response["detail"]
                or "authorized" in response["detail"]
            ):
                raise UnauthorizedError(
                    f"Unauthorized Tiingo request -> {response['detail']}"
                )
            raise OpenBBError(response["detail"])

        if not response:
            raise EmptyDataError("The response is empty")

        return response

    except JSONDecodeError as e:
        raise OpenBBError(f"Failed to parse JSON response -> {e}") from e



"""Tiingo Company News Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Dict, List, Optional

from dateutil import parser
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.company_news import (
    CompanyNewsData,
    CompanyNewsQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field, field_validator


class TiingoCompanyNewsQueryParams(CompanyNewsQueryParams):
    """Tiingo Company News Query.

    Source: https://www.tiingo.com/documentation/news
    """

    __alias_dict__ = {
        "symbol": "tickers",
        "start_date": "startDate",
        "end_date": "endDate",
    }
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
        "source": {"multiple_items_allowed": True},
    }

    offset: Optional[int] = Field(
        default=0, description="Page offset, used in conjunction with limit."
    )
    source: Optional[str] = Field(
        default=None, description="A comma-separated list of the domains requested."
    )


class TiingoCompanyNewsData(CompanyNewsData):
    """Tiingo Company News data."""

    __alias_dict__ = {
        "symbols": "tickers",
        "date": "publishedDate",
        "text": "description",
        "article_id": "id",
    }

    tags: Optional[str] = Field(
        default=None, description="Tags associated with the news article."
    )
    article_id: int = Field(description="Unique ID of the news article.")
    source: str = Field(description="News source.")
    crawl_date: datetime = Field(description="Date the news article was crawled.")

    @field_validator("tags", "symbols", mode="before")
    @classmethod
    def list_to_string(cls, v):
        """Convert list to string."""
        return ",".join(v) if v else None

    @field_validator("crawl_date", mode="before")
    @classmethod
    def validate_date(cls, v):
        """Validate the date."""
        return parser.parse(v)


class TiingoCompanyNewsFetcher(
    Fetcher[
        TiingoCompanyNewsQueryParams,
        List[TiingoCompanyNewsData],
    ]
):
    """Transform the query, extract and transform the data from the Tiingo endpoints."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> TiingoCompanyNewsQueryParams:
        """Transform the query params."""
        return TiingoCompanyNewsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: TiingoCompanyNewsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the tiingo endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        import math
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_tiingo.utils.helpers import get_data

        api_key = credentials.get("tiingo_token") if credentials else ""

        base_url = "https://api.tiingo.com/tiingo/news"
        query_str = get_querystring(
            query.model_dump(by_alias=False), ["limit", "offset"]
        )

        limit = query.limit if query.limit else 1000
        pages = 0
        if limit > 1000:
            pages = math.ceil(limit / 1000)
            limit = 1000
            urls = [
                f"{base_url}?{query_str}&token={api_key}&limit={limit}&offset={page * 1000 if page > 0 else 0}"
                for page in range(0, pages)
            ]
        else:
            urls = [f"{base_url}?{query_str}&token={api_key}&limit={limit}"]

        results: list = []

        async def get_one(url):
            """Get data for one URL and append results to list."""
            response = await get_data(url)
            if isinstance(response, list):
                results.extend(response)
            elif isinstance(response, dict):
                results.append(response)

        await asyncio.gather(*[get_one(url) for url in urls])

        if not results:
            raise EmptyDataError()

        return results

    @staticmethod
    def transform_data(
        query: TiingoCompanyNewsQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[TiingoCompanyNewsData]:
        """Return the transformed data."""
        return [TiingoCompanyNewsData.model_validate(d) for d in data[: query.limit]]



"""Tiingo Equity Historical Price Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal, Optional, Union
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_historical import (
    EquityHistoricalData,
    EquityHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
from pydantic import Field


class TiingoEquityHistoricalQueryParams(EquityHistoricalQueryParams):
    """Tiingo Equity Historical Price Query.

    Source: https://www.tiingo.com/documentation/end-of-day
    """

    __alias_dict__ = {
        "start_date": "startDate",
        "end_date": "endDate",
    }
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
        "interval": {
            "choices": [
                "1m",
                "5m",
                "15m",
                "30m",
                "90m",
                "1h",
                "2h",
                "4h",
                "1d",
                "1W",
                "1M",
                "1Y",
            ]
        },
    }

    interval: Union[
        Literal[
            "1m", "5m", "15m", "30m", "90m", "1h", "2h", "4h", "1d", "1W", "1M", "1Y"
        ],
        str,
    ] = Field(default="1d", description=QUERY_DESCRIPTIONS.get("interval", ""))


class TiingoEquityHistoricalData(EquityHistoricalData):
    """Tiingo Equity Historical Price Data."""

    __alias_dict__ = {
        "adj_open": "adjOpen",
        "adj_high": "adjHigh",
        "adj_low": "adjLow",
        "adj_close": "adjClose",
        "adj_volume": "adjVolume",
        "split_ratio": "splitFactor",
        "dividend": "divCash",
    }

    adj_open: Optional[float] = Field(
        default=None,
        description="The adjusted open price.",
    )
    adj_high: Optional[float] = Field(
        default=None,
        description="The adjusted high price.",
    )
    adj_low: Optional[float] = Field(
        default=None,
        description="The adjusted low price.",
    )
    adj_close: Optional[float] = Field(
        default=None,
        description=DATA_DESCRIPTIONS.get("adj_close", ""),
    )
    adj_volume: Optional[float] = Field(
        default=None,
        description="The adjusted volume.",
    )
    split_ratio: Optional[float] = Field(
        default=None,
        description="Ratio of the equity split, if a split occurred.",
    )
    dividend: Optional[float] = Field(
        default=None,
        description="Dividend amount, if a dividend was paid.",
    )


class TiingoEquityHistoricalFetcher(
    Fetcher[
        TiingoEquityHistoricalQueryParams,
        list[TiingoEquityHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the Tiingo endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> TiingoEquityHistoricalQueryParams:
        """Transform the query params."""
        # pylint: disable=import-outside-toplevel
        from dateutil.relativedelta import relativedelta

        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return TiingoEquityHistoricalQueryParams(**transformed_params)

    # pylint: disable=protected-access
    @staticmethod
    async def aextract_data(
        query: TiingoEquityHistoricalQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the Tiingo endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_tiingo.utils.helpers import get_data

        api_key = credentials.get("tiingo_token") if credentials else ""

        base_url = (
            "https://api.tiingo.com/tiingo/daily"
            if query.interval in ["1d", "1W", "1M", "1Y"]
            else "https://api.tiingo.com/iex"
        )
        query_str = get_querystring(
            query.model_dump(by_alias=True), ["symbol", "interval"]
        )
        frequency_dict = {
            "1d": "daily",
            "1W": "weekly",
            "1M": "monthly",
            "1Y": "annually",
        }
        frequency = (
            frequency_dict.get(query.interval, "")
            if query.interval in frequency_dict
            else query.interval
        )
        cols_str = "&columns=open,high,low,close,volume"

        if frequency.endswith("m"):
            frequency = f"{frequency[:-1]}min"
            query_str = query_str + cols_str
        elif frequency == "h":
            frequency = f"{frequency[:-1]}hour"
            query_str = query_str + cols_str

        results: list = []
        messages: list = []

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}/{symbol}/prices?{query_str}&resampleFreq={frequency}&token={api_key}"
            data = None
            try:
                data = await get_data(url)
            except UnauthorizedError as e:
                raise e from e
            except OpenBBError as e:
                if (
                    e.original
                    and isinstance(e.original, str)
                    and "ticker" in e.original.lower()
                ):
                    messages.append(e.original)
                else:
                    messages.append(f"{symbol}: {e.original}")

            if isinstance(data, list):
                if "," in query.symbol:
                    for d in data:
                        d["symbol"] = symbol
                results.extend(data)

        symbols = query.symbol.split(",")
        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results and messages:
            raise OpenBBError(f"{messages}")

        if not results and not messages:
            raise EmptyDataError("The request was returned empty.")

        if results and messages:
            for message in messages:
                warn(message)

        return results

    @staticmethod
    def transform_data(
        query: TiingoEquityHistoricalQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[TiingoEquityHistoricalData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from pandas import to_datetime
        from pytz import timezone

        results: list[TiingoEquityHistoricalData] = []
        for item in data:
            # If the data is intraday, it comes from IEX and the TZ should be localized to America/New_York
            if query.interval in ["1d", "1W", "1M", "1Y"]:
                item["date"] = to_datetime(item["date"]).date()
            else:
                item["date"] = to_datetime(item["date"], utc=True).tz_convert(
                    timezone("America/New_York")
                )

            results.append(TiingoEquityHistoricalData.model_validate(item))

        return results



"""Tiingo Currency Historical Price Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal, Optional, Union
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.currency_historical import (
    CurrencyHistoricalData,
    CurrencyHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
from pydantic import Field


class TiingoCurrencyHistoricalQueryParams(CurrencyHistoricalQueryParams):
    """Tiingo Currency Historical Price Query.

    Source: https://www.tiingo.com/documentation/forex
    """

    __alias_dict__ = {
        "start_date": "startDate",
        "end_date": "endDate",
    }
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
        "interval": {
            "choices": [
                "1m",
                "5m",
                "15m",
                "30m",
                "90m",
                "1h",
                "2h",
                "4h",
                "1d",
                "5d",
                "21d",
            ]
        },
    }

    interval: Union[
        Literal[
            "1m",
            "5m",
            "15m",
            "30m",
            "90m",
            "1h",
            "2h",
            "4h",
            "1d",
            "5d",
            "21d",
        ],
        str,
    ] = Field(default="1d", description=QUERY_DESCRIPTIONS.get("interval", ""))


class TiingoCurrencyHistoricalData(CurrencyHistoricalData):
    """Tiingo Currency Historical Price Data."""

    __alias_dict__ = {"symbol": "ticker"}


class TiingoCurrencyHistoricalFetcher(
    Fetcher[
        TiingoCurrencyHistoricalQueryParams,
        list[TiingoCurrencyHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the Tiingo endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> TiingoCurrencyHistoricalQueryParams:
        """Transform the query params."""
        # pylint: disable=import-outside-toplevel
        from dateutil.relativedelta import relativedelta

        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return TiingoCurrencyHistoricalQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: TiingoCurrencyHistoricalQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Any,
    ) -> list[dict]:
        """Return the raw data from the Tiingo endpoint."""
        # pylint: disable=import-outside-toplevel
        import asyncio  # noqa
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_tiingo.utils.helpers import get_data
        from pandas import to_datetime

        api_key = credentials.get("tiingo_token") if credentials else ""
        base_url = "https://api.tiingo.com/tiingo/fx"
        query_str = get_querystring(
            query.model_dump(by_alias=False), ["tickers", "resampleFreq"]
        )

        query_str = get_querystring(
            query.model_dump(by_alias=True), ["symbol", "interval"]
        )

        if query.interval.endswith("m"):
            frequency = f"{query.interval[:-1]}min"
        elif query.interval.endswith("h"):
            frequency = f"{query.interval[:-1]}hour"
        elif query.interval.endswith("d"):
            frequency = f"{query.interval[:-1]}day"
        else:
            frequency = "1day"

        results: list = []
        messages: list = []
        symbols = query.symbol.split(",")

        async def get_one(symbol):
            """Get data for one symbol."""
            url = f"{base_url}/{symbol}/prices?{query_str}&resampleFreq={frequency}&token={api_key}"
            data = None
            try:
                data = await get_data(url)
            except UnauthorizedError as e:
                raise e from e
            except OpenBBError as e:
                if (
                    e.original
                    and isinstance(e.original, str)
                    and "ticker" in e.original.lower()
                ):
                    messages.append(e.original)
                else:
                    messages.append(f"{symbol}: {e.original}")

            if isinstance(data, list):
                for d in data:
                    ticker = d.pop("ticker", None)
                    if ticker and len(symbols) > 1:
                        d["ticker"] = d["ticker"].upper()

                    if query.interval.endswith("d"):
                        d["date"] = to_datetime(d["date"]).date()
                    else:
                        d["date"] = to_datetime(d["date"], utc=True)

                results.extend(data)

        await asyncio.gather(*[get_one(symbol) for symbol in symbols])

        if not results and messages:
            raise OpenBBError(f"{messages}")

        if not results and not messages:
            raise EmptyDataError("The request was returned empty.")

        if results and messages:
            for message in messages:
                warn(message)

        return results

    @staticmethod
    def transform_data(
        query: TiingoCurrencyHistoricalQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[TiingoCurrencyHistoricalData]:
        """Return the transformed data."""
        return sorted(
            [TiingoCurrencyHistoricalData.model_validate(d) for d in data],
            key=lambda x: x.date,
        )



"""Tiingo Crypto Historical Price Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Literal, Optional, Union
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.crypto_historical import (
    CryptoHistoricalData,
    CryptoHistoricalQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError, UnauthorizedError
from pydantic import Field


class TiingoCryptoHistoricalQueryParams(CryptoHistoricalQueryParams):
    """Tiingo Crypto Historical Price Query.

    Source: https://www.tiingo.com/documentation/crypto
    """

    __alias_dict__ = {
        "start_date": "startDate",
        "end_date": "endDate",
        "symbol": "tickers",
    }
    __json_schema_extra__ = {
        "symbol": {"multiple_items_allowed": True},
        "exchanges": {"multiple_items_allowed": True},
        "interval": {
            "choices": [
                "1m",
                "5m",
                "15m",
                "30m",
                "90m",
                "1h",
                "2h",
                "4h",
                "1d",
                "7d",
                "30d",
            ]
        },
    }

    interval: Union[
        Literal[
            "1m",
            "5m",
            "15m",
            "30m",
            "90m",
            "1h",
            "2h",
            "4h",
            "1d",
            "7d",
            "30d",
        ],
        str,
    ] = Field(default="1d", description=QUERY_DESCRIPTIONS.get("interval", ""))

    exchanges: Optional[Union[list[str], str]] = Field(
        default=None,
        description=(
            "To limit the query to a subset of exchanges e.g. ['POLONIEX', 'GDAX']"
        ),
    )


class TiingoCryptoHistoricalData(CryptoHistoricalData):
    """Tiingo Crypto Historical Price Data."""

    __alias_dict__ = {
        "transactions": "tradesDone",
        "volume_notional": "volumeNotional",
    }

    transactions: Optional[int] = Field(
        default=None,
        description="Number of transactions for the symbol in the time period.",
    )

    volume_notional: Optional[float] = Field(
        default=None,
        description=(
            "The last size done for the asset on the specific date in the "
            "quote currency. The volume of the asset on the specific date in "
            "the quote currency."
        ),
    )


class TiingoCryptoHistoricalFetcher(
    Fetcher[
        TiingoCryptoHistoricalQueryParams,
        list[TiingoCryptoHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the Tiingo endpoints."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> TiingoCryptoHistoricalQueryParams:
        """Transform the query params."""
        # pylint: disable=import-outside-toplevel
        from dateutil.relativedelta import relativedelta

        transformed_params = params

        now = datetime.now().date()
        if params.get("start_date") is None:
            transformed_params["start_date"] = now - relativedelta(years=1)

        if params.get("end_date") is None:
            transformed_params["end_date"] = now

        return TiingoCryptoHistoricalQueryParams(**transformed_params)

    @staticmethod
    async def aextract_data(
        query: TiingoCryptoHistoricalQueryParams,
        credentials: Optional[dict[str, str]],
        **kwargs: Optional[Any],
    ) -> list[dict]:
        """Return the raw data from the Tiingo endpoint."""
        # pylint: disable=import-outside-toplevel
        from openbb_core.provider.utils.helpers import get_querystring
        from openbb_tiingo.utils.helpers import get_data

        api_key = credentials.get("tiingo_token") if credentials else ""

        base_url = "https://api.tiingo.com/tiingo/crypto/prices"
        query_str = get_querystring(query.model_dump(by_alias=True), ["interval"])

        if query.interval.endswith("m"):
            frequency = f"{query.interval[:-1]}min"
        elif query.interval.endswith("h"):
            frequency = f"{query.interval[:-1]}hour"
        elif query.interval.endswith("d"):
            frequency = f"{query.interval[:-1]}day"
        else:
            frequency = "1day"

        results: list = []
        url = f"{base_url}?{query_str}&resampleFreq={frequency}&token={api_key}"

        try:
            results = await get_data(url)
        except (EmptyDataError, OpenBBError, UnauthorizedError) as e:
            raise e from e

        return results

    @staticmethod
    def transform_data(
        query: TiingoCryptoHistoricalQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[TiingoCryptoHistoricalData]:
        """Return the transformed data."""
        # pylint: disable=import-outside-toplevel
        from pandas import to_datetime

        results: list[TiingoCryptoHistoricalData] = []
        symbols = query.symbol.split(",")
        returned_symbols = [item.get("ticker", "").upper() for item in data]

        for symbol in symbols:
            if symbol not in returned_symbols:
                warn(f"No data found for {symbol}")

        for item in data:
            symbol = item.get("ticker", "").upper()
            price_data = item.get("priceData", [])

            if not price_data:
                warn(f"No data found for {symbol}")
                continue

            for row in price_data:
                if len(returned_symbols) > 1:
                    row["symbol"] = symbol
                if query.interval.endswith("d"):
                    row["date"] = to_datetime(row["date"]).date()
                else:
                    row["date"] = to_datetime(row["date"], utc=True)

                results.append(TiingoCryptoHistoricalData.model_validate(row))

        return sorted(results, key=lambda x: x.date)
